from urllib.parse import quote as urlquote


def make_dsn(protocol: str, *, host: str, port: int, database: str, username: str, password: str) -> str:
    username = urlquote(username, safe="")
    password = urlquote(password, safe="")

    return f"{protocol}://{username}:{password}@{host}:{port}/{database}"


def make_postgres_dsn(*, host: str, database: str, username: str, password: str, port: int = 5432) -> str:
    return make_dsn(
        protocol='postgresql',
        host=host,
        port=port,
        database=database,
        username=username,
        password=password,
    )



from __future__ import annotations

import time
import typing as ty


class Timer:
    def __init__(self, parent: Stopwatch, name: str):
        self._parent = parent
        self._name = name
        self._entered: float = 0

    def __enter__(self):
        self._entered = time.monotonic()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        d = time.monotonic() - self._entered
        self._parent[self._name] = d


class Stopwatch:
    def __init__(self):
        self._started: float = time.monotonic()
        self._times: ty.Dict[str, float] = {}

    @property
    def times(self):
        return {**self._times, 'total': time.monotonic() - self._started}

    def __getitem__(self, item):
        return Timer(self, item)

    def __setitem__(self, name: str, duration: float):
        self._times[name] = duration



from itertools import takewhile, islice, repeat
from typing import TypeVar, Dict, List, Iterable, Callable, Iterator

U = TypeVar('U')
R = TypeVar('R')


def grouper(iterable: Iterable[U], n) -> Iterator[List[U]]:
    it = iter(iterable)
    yield from takewhile(bool, ([*islice(it, n)] for _ in repeat(None)))


def groupby(f: Callable[[U], R], xs: Iterable[U]) -> Dict[R, List[U]]:
    r = dict()
    for x in xs:
        k = f(x)
        if k not in r:
            r[k] = []
        r[k].append(x)
    return r


def dedup(xs: Iterable[U], on=lambda x: x) -> Iterator[U]:
    """
    Deduplicates an iterable. Any item is guaranteed to only occur once in the result.
    """

    seen = set()
    for x in xs:
        identity = on(x)
        if identity in seen:
            continue
        seen.add(identity)
        yield x



from .base_convert import \
    base_convert
from .daterange import \
    DateRange
from .dates import \
    at_end_of_month, \
    read_date
from .db import \
    make_dsn, \
    make_postgres_dsn
from .iterables import \
    grouper, \
    groupby, \
    dedup
from .ranges import \
    range_incl, \
    range_excl
from .singleton import \
    Singleton
from .stopwatch import \
    Stopwatch

__all__ = [
    'base_convert',
    'DateRange',
    'at_end_of_month',
    'read_date',
    'make_dsn',
    'make_postgres_dsn',
    'grouper',
    'groupby',
    'dedup',
    'range_incl',
    'range_excl',
    'Singleton',
    'Stopwatch',
]



def base_convert(s, from_base: int, to_base: int, alphabet="0123456789abcdefghijklmnopqrstuvwxyz") -> str:
    s = str(s)
    n = 0
    p = 1
    for digit in s[::-1]:
        if digit in alphabet:
            n += alphabet.index(digit) * p
            p *= from_base
    if n == 0:
        return '0'
    digits = []
    while n:
        digits.append(alphabet[n % to_base])
        n //= to_base
    return ''.join(digits[::-1])



from datetime import date, timedelta
from numbers import Number
from typing import TypeVar, Optional, Union, Callable, Iterator, Type

E = TypeVar('E', Number, date, covariant=True)


def _range(lower: E, step: Callable[[E], E], condition: Callable[[E], bool]) -> Iterator[E]:
    """
    Very generic range function. Yields a stream from lower, incremented by the given step-function,
    until the given condition is reached.

    spec: takeWhile predicate (iterate step seed)
    """
    current = lower
    while condition(current):
        yield current
        current = step(current)


def _get_incrementor(type_: Type[E], step: Number = 1) -> Callable[[E], E]:
    """
    Returns a function that increments values of the given type by the given step.

    ...basically `(+ 1)` (or `(+ step)` respectively), but can increment other types as well. Currently
    the only other type supported is datetime.date which is incremented in days.
    """
    if issubclass(type_, date):
        # noinspection PyTypeChecker
        return lambda value: value + timedelta(days=step)
    return lambda value: value + step


def range_excl(lower: E, upper: E, step: Optional[Union[Callable[[E], E], Number]] = None) -> Iterator[E]:
    if step is None:
        step = _get_incrementor(type(lower))
    elif isinstance(step, Number):
        step = _get_incrementor(type(lower), step)
    yield from _range(lower, step, lambda value: value < upper)


def range_incl(lower: E, upper: E, step: Optional[Union[Callable[[E], E], Number]] = None) -> Iterator[E]:
    if step is None:
        step = _get_incrementor(type(lower))
    elif isinstance(step, Number):
        step = _get_incrementor(type(lower), step)
    yield from _range(lower, step, lambda value: value <= upper)



import re

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

    start_result = re.search(r"<!-- START doctoc [^>]+-->", long_description)
    if start_result:
        pos = start_result.span(0)[0]
        before = long_description[:pos]
        remaining = long_description[pos:]
        end_result = re.search(r"<!-- END doctoc [^>]+-->", remaining)
        end_pos = end_result.span(0)[1]
        after = remaining[end_pos:]
        long_description = before + after

__pkginfo__ = {}
with open("allthethings/__pkginfo__.py") as fh:
    exec(fh.read(), __pkginfo__)

setuptools.setup(
    name="allthethings",
    version=__pkginfo__['__version__'],
    author="Julian Fleischer",
    author_email="tirednesscankill@warhog.net",
    description="Various utilities",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/scravy/allthethings",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import unittest
from dataclasses import dataclass

from allthethings.iterables import dedup, grouper


class IterablesTestCase(unittest.TestCase):
    def test_grouper(self):
        result = [*grouper([1, 2, 3, 4, 5, 6, 7], n=3)]
        self.assertEqual([[1, 2, 3], [4, 5, 6], [7]], result)

    def test_dedup(self):
        result = [*dedup([1, 2, 3, 2, 1])]
        self.assertEqual([1, 2, 3], result)

    def test_dedup_on(self):
        @dataclass
        class X:
            a: str
            b: str

        xs = [X("one", "foo"), X("one", "bar"), X("two", "baz"), X("one", "qux"), X("three", "baz")]
        self.assertEqual([X("one", "foo"), X("two", "baz"), X("three", "baz")], list(dedup(xs, on=lambda x: x.a)))


if __name__ == '__main__':
    unittest.main()
